! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: slconv_mod
!
! Authors:  Julian Osborne - SSC/LUX
!           Ian Stewart - SSC/LUX
!           Maite Ceballos - SSC/IFCA
!           Masaaki Sakano - SSC/LUX
!
! Description: Converts XMM ML or summary source lists to gaia or ds9 format.
! 
! TODO:
!  * Consider clobber()
!
! $Id: slconv_mod.f90,v 2.0 2005/02/07 09:59  Exp $
!- 
module slconv_mod

  use dal
  use param  
  use errorHandling
  use datafilter
  use exprcolgen	! in selectlib

  use array_utils
  use constants
  use coordinate	! for raDecToTan
  use dal_aux		! for splitSetTabName
  use sort_mod
  use ssc_misc
  use warp_utils	! for scaleBalance
  use wcs_aux

  implicit none 
  
  private 

  integer, parameter :: lenGaiaColNames = 20

  ! Command-line parameters:
  character(255) ::&
    tempSetName = '',&
    srcListSetTabName = '',&
    imageSetName = '',&
    selectionExpression = '',&
    maxSourceNumSortExpression = '',&
    srcLabelSortExpression = '',&
    srcLabelExpression = '',&
    radiusExpression = '',&
    outFileRoot = '',&
    outFileName = ''
  character(20) ::&
    colour = '',&
    shape = '',&
    radiusUnit = '',&
    radiusStyle = '',&
    srcLabelStyle = '',&
    srcLabelSortStyle = '',&
    raColName = '',&
    decColName = '',&
    raDecUnits = '',&
    outFileStyle = ''
  character(lenGaiaColNames), allocatable ::&
    extraGaiaColNames(:)
  character(4) ::&
    outFileFormat = ''
  integer(int32) ::&
    maxSourceNum = 0
  logical ::&
    withSrcLabels = .false.,&
    truncateSourceList = .false.,&
    withImageSet = .false.

  ! Other globals:
  type :: localSrcDataType
!    integer(int32) ::&
!      index
    real(double) ::&
      raDeg,&
      decDeg
    real(single) ::&
      xPixels,&
      yPixels,&
      radius
    character(20) ::&
      label
  end type localSrcDataType

  integer, parameter :: numMandatoryGaiaCols = 4
  character(255) ::&
    srcListSetname
  character(20) ::&
    srcListTabName
  character(lenGaiaColNames) ::&
    mandatoryGaiaColNames(numMandatoryGaiaCols)&
    = (/'Id    ', 'ra    ', 'dec   ', 'radius'/)
  real(single), parameter ::&
    minRadiusFraction = 0.005,&
    maxRadiusFraction = 0.025,&
    dragMedianTo = 0.3,&
    lastResortFovDiameterArcsec = 1800.0,&
    textYOffsetFrac = 0.01
  type(WcsT) ::&
    wcs
  real(single) :: maxSize ! either in pixels or arcsec.

  public slconv

contains 

  !---------------------------------------------------------------------
  subroutine slconv
  !---------------------------------------------------------------------
  ! - Converts any kind of FITS source list to gaia or ds9 format.

    implicit none

    type(localSrcDataType), pointer :: srcData(:)
    type(DataSetT) :: imageSet, sourcesSet
    type(ArrayT) :: imageArray

    !-------------------------------------------------------------------
    ! Start

    call readParameters
    if (withImageSet) then
      imageSet = dataSet(imageSetName, READ)
      imageArray = array(imageSet, 0, READ)
      call readWcs(imageSet, imageArray, wcs)
      call release(imageSet)
    end if

    call filterSourcesAndLoadSomeData(srcData, sourcesSet)

    if (size(srcData)>0) then
      if (ucase(outFileFormat) .eq. 'DS9') then
        call saveAsDs9RegionFile(srcData)
      else if (ucase(outFileFormat) .eq. 'GAIA') then
        call saveAsGaiaRegionFile(srcData, sourcesSet)
      else
call error('badOutputStyle', "The value of the --outputstyle parameter &
&was not recognized.")
      end if
    end if
    deallocate(srcData)
    call release(sourcesSet)

  end subroutine slconv
  
  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: i, numExtraGaiaColNames

    !-------------------------------------------------------------------
    ! Start

!    tempSetName = stringParameter('tempset')
tempSetName = 'tempset.ds'
!*** get rid of the parameter.

    srcListSetTabName = stringParameter('srclisttab')
    raColName  = stringParameter('racolumn')
    decColName = stringParameter('deccolumn')
    raDecUnits = stringParameter('radecunits')
    selectionExpression = stringParameter('expression')

    truncateSourceList = booleanParameter('withncut')
    if (truncateSourceList) then
      maxSourceNumSortExpression = stringParameter('ncutsortexpression')
      maxSourceNum = intParameter('ncut')
    end if

!    withImageSet = booleanParameter('withimageset')
!    imageSetName = stringParameter('imageset')
withImageSet = .false.

!    radiusUnit       = stringParameter('radiusunit')
    if (.not. withImageSet) then
      ! This is ALWAYS the case (see immediately above). 
      ! As a result, ALWAYS radiusUnit=='arcsec'
      ! Note the statememt
      !    if (radiusUnit .eq. 'pixels') then
      ! appears from place to place in this code!!  Check!!!
      radiusUnit = 'arcsec'
    end if

    radiusExpression = stringParameter('radiusexpression')
    radiusStyle      = stringParameter('radiusstyle')

    withSrcLabels = booleanParameter('withlabels')
    if (withSrcLabels) then
      srcLabelStyle = stringParameter('labelstyle')
      select case(srcLabelStyle)
      case('sortedint')
        srcLabelSortStyle  = stringParameter('labelsortstyle')
        select case(srcLabelStyle)
        case('radius')
        case('rownumber')
        case('expr')
          srcLabelSortExpression = stringParameter('sortexpression')
        case default
call error('badLabelSortStyle',"The value of the --labelsortstyle &
&parameter was not recognized.")
        end select

      case('expr')
        srcLabelExpression = stringParameter('labelexpression')

      case default
call error('badLabelStyle',"The value of the --labelstyle parameter &
&was not recognized.")
      end select

!      colour = stringParameter('labelcolour')
    end if

colour = stringParameter('colour')

shape = stringParameter('shape')

    outFileFormat = stringParameter('outputstyle')
    if (ucase(outFileFormat) .eq. 'GAIA') then
      numExtraGaiaColNames = parameterCount('extragaiacolumns')
      allocate(extraGaiaColNames(numExtraGaiaColNames))
      do i = 0, numExtraGaiaColNames - 1
        extraGaiaColNames(i + 1) = stringParameter('extragaiacolumns', i)
      end do
    end if

    outFileStyle = stringParameter('outfilestyle')
    select case(outFileStyle)
    case('prefix')
      outFileRoot = stringParameter('outfileprefix')
    case('whole')
      outFileName = stringParameter('outfile')
    case default
call error('badOutFileStyle',"The value of the --outfilestyle parameter &
&was not recognized.")
    end select

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine filterSourcesAndLoadSomeData(srcData, sourcesSet)
  !---------------------------------------------------------------------

    implicit none

    type(localSrcDataType), pointer :: srcData(:)
    type(DataSetT), intent(out) :: sourcesSet

    type(TableT) :: sourcesTab
    type(ColumnT) :: labelStringCol, indexCol
    real(double), allocatable :: raColData(:), decColData(:), sortColData(:)
    integer(int32) ::&
      numFilteredRows,&
      i
    real(single), allocatable ::&
      tempRadii(:)
    character(20) ::&
      nCutSortColName  = '',&
      nCutIndexColName = '',&
      radiusColName    = '',&
      labelSortColName = '',&
!      labelExprColName = '',&
      maxSrcNumStr
!    logical(bool) :: someRadiiAreNegative, someRadiiAreNulls, allRadiiAreNulls
    logical :: someRadiiAreNegative, someRadiiAreNulls, allRadiiAreNulls
    real(single) ::&
      maxRadiusDataValue,&
      minRadius, maxRadius
    integer(int32), pointer ::&
      tempIndices(:)
    integer ::&
      foreignStatus
    integer(int32), allocatable :: int32ColData(:)
    integer(int32), pointer :: indexData(:)
    real(single),   allocatable :: real32ColData(:)
    real(double),   allocatable :: real64ColData(:)

  !-------------------------------------------------------------------
  ! Start

    call message(AppMsg, Verbose, 'Open source list.')

    !...................................................................
    ! Load and filter source list:

    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)
    ! in ssclib/dal_aux.

    sourcesSet = clone(srcListSetName, tempSetName, TEMP)
    sourcesTab = table(sourcesSet, srcListTabName, MODIFY)

    call message(AppMsg, Verbose, 'Selection of sources.')

    call initFilter(selectionExpression)
    numFilteredRows = doFilter(sourcesTab)

    if (numFilteredRows .le. 0) then
      call warning('noSourcesSelected'&
        , 'No sources passed the specified filtering expression.')
      allocate(srcData(0))
  return
    end if

    !...................................................................
    ! If a limit on the number of sources to be plotted is desired, the sources
    ! must first be sorted. This is accomplished next:

    if (truncateSourceList) then
      call message(AppMsg, Verbose, 'Sorting to prepare for max source &
      &number cut.')

!nCutSortColName = generateNewColName(sourcesTab) ! in ssclib/dal_aux.
nCutSortColName = 'T_RESULT'
if (hasColumn(sourcesTab, nCutSortColName)) then
!*** delete this column
call error('tResultColAlreadyExists', "slconv wants to make a column &
&in the source list to hold the result of its column arithmetic. &
&However it is not yet smart enough to choose an unused column name &
&- it insists on it being called T_RESULT. However there is already &
&a column of that name in the file.")
end if

call initExprColGen(maxSourceNumSortExpression, nCutSortColName, REAL32)
call doExprColGen(sourcesTab, sourcesTab)

!nCutIndexColName = generateNewColName(sourcesTab, (/nCutSortColName/)) ! in ssclib/dal_aux.
nCutIndexColName = 'INDEX'
if (hasColumn(sourcesTab, nCutIndexColName)) then
!*** delete this column
call error('indexColAlreadyExists', "slconv wants to make a column &
&in the source list to hold a sort index. However it is not yet &
&smart enough to choose an unused column name - it insists on it &
&being called INDEX. However there is already a column of that name &
&in the file.")
end if


      indexCol = addColumn(sourcesTab, nCutIndexColName, INTEGER32)
      indexData => int32Data(indexCol)

      allocate(sortColData(numFilteredRows))
      call readColDataToFixed(sourcesTab, nCutSortColName, sortColData)
      sortColData = -sortColData
      allocate(tempIndices(numFilteredRows))
      tempIndices = (/(i, i=1,numFilteredRows)/)

      call quickSort(sortColData, tempIndices)
      do i = 1, numFilteredRows
        indexData(tempIndices(i) - 1) = i
      end do
      deallocate(sortColData, tempIndices)
      call release(indexCol)

      maxSrcNumStr = intToStr(maxSourceNum)
      write(selectionExpression, '(a)') "INDEX<="//trim(maxSrcNumStr)
      call initFilter(selectionExpression)
      numFilteredRows = doFilter(sourcesTab)
    end if

    !...................................................................

    allocate(srcData(numFilteredRows), raColData(numFilteredRows))
    allocate(decColData(numFilteredRows))

    call readColDataToFixed(sourcesTab, raColName,  raColData)
    call readColDataToFixed(sourcesTab, decColName, decColData)

!**** check ra/dec unit against any that is recorded in the file??

    select case(raDecUnits)
    case('decimaldeg')
      srcData%raDeg  = raColData
      srcData%decDeg = decColData
    case('hdms')
call error('notYetSupported',"Sorry, this mode is not yet supported.")
    case('radians')
      srcData%raDeg  = raColData  / PiOn180
      srcData%decDeg = decColData / PiOn180
    case default
call error('badRaDecUnit', "The value of the --radecunits parameter &
&was not recognized.")
    end select

    deallocate(raColData)
    deallocate(decColData)

    !...................................................................
    ! Calculate radii of source circles:

    call message(AppMsg, Verbose, 'Calculating radii.')

!radiusColName = generateNewColName(sourcesTab, (/nCutSortColName, nCutIndexColName/)) ! in ssclib/dal_aux.
radiusColName = 'R_RESULT'
if (hasColumn(sourcesTab, radiusColName)) then
!*** delete this column
call error('rResultColAlreadyExists', "slconv wants to make a column &
&in the source list to hold the result of its column arithmetic. &
&However it is not yet smart enough to choose an unused column name &
&- it insists on it being called R_RESULT. However there is already &
&a column of that name in the file.")
end if

    call initExprColGen(radiusExpression, radiusColName, REAL32)
    call doExprColGen(sourcesTab, sourcesTab)

    call maxNonNullValue(sourcesTab, radiusColName, maxRadiusDataValue&
      , allRadiiAreNulls) ! in ssclib/dal_aux

    if (allRadiiAreNulls)&
call error('allRadiiAreNulls', 'All the rows specified by &
&--radiusexpression are null-valued.')

    call readColDataToFixed(sourcesTab, radiusColName, srcData%radius)

    someRadiiAreNegative = .false. ! default
    someRadiiAreNulls    = .false. ! default
    do i = 1, numFilteredRows
      write(errstr, '(a, i4, a, e10.3)') 'Source '&
        , i, ', radius = ', srcData(i)%radius
      call message(AppMsg, Noisy)

      if (srcData(i)%radius < 0.0)  someRadiiAreNegative = .true.
      if (isNaN(srcData(i)%radius)) someRadiiAreNulls = .true.
    end do

    if (someRadiiAreNegative) then
      call warning('someRadiiAreNegative',"Some of the source radii were &
        &found to be less than zero. These values will be replaced by zero.")
    end if

    if (someRadiiAreNulls) then
      call warning('someRadiiAreNulls',"Some of the source radii were found  &
        &to be NaN. Such sources won't be plotted.")
    end if

    !...................................................................
    ! Adjust radii of source circles, if desired:

    call calcMainScale(srcData)
    if (radiusStyle .eq. 'auto') then
      call scaleBalance(srcData%radius, dragMedianTo, foreignStatus)

      if (foreignStatus < 0) then
        select case(foreignStatus)
        case(-1)
          call warning('radiiAllEqual',"The maximum and minimum source radii are &
          &the same. This can happen in a number of ways. Maybe there is only one &
          &source; or maybe you have a constant expression in --radiusexpression. &
          &In the latter case there is no point in asking for --radiusstyle=auto.")

          minRadius = minRadiusFraction * maxSize
          maxRadius = maxRadiusFraction * maxSize
          srcData%radius = (minRadius + maxRadius) / 2.0
        case(-2)
          call warning('noSources',"No sources have passed the filtering expression.")
        case default
call error('scaleBalanceError',"The subroutine &
&scaleBalance returned a non-zero error status. Contact the code &
&developer please.")
        end select

      else
        minRadius = minRadiusFraction * maxSize
        maxRadius = maxRadiusFraction * maxSize

        do i = 1, numFilteredRows
          srcData(i)%radius = fracVal(minRadius, maxRadius, srcData(i)%radius)
!********** not sure that this is doing the right thing - doesn't this assume that the srcData(i)%radius all come out of scaleBalance with values between 0 and 1?
        end do
      end if
    end if ! radiusStyle .eq. 'auto'

    !...................................................................
    ! Allocating source labels:

    srcData%label = ''

    if (withSrcLabels) then
      call message(AppMsg, Verbose, 'Allocating source labels.')

      allocate(tempRadii(numFilteredRows))

      select case(srcLabelStyle)
      case('sortedint')
        select case(srcLabelSortStyle)

!        case('radiusup')
!          tempRadii = -srcData%radius
!!          tempIndices = indicize(numFilteredRows)
!tempIndices => indicize(numFilteredRows)
!          call quickSort(tempRadii, tempIndices)
!          do i = 1, numFilteredRows
!            srcData(tempIndices(i))%label = intToStr(i)
!          end do
!
!        case('radiusdown')

        case('radius')
          tempRadii = -srcData%radius
          allocate(tempIndices(numFilteredRows))
          tempIndices = (/(i, i=1,numFilteredRows)/)
          call quickSort(tempRadii, tempIndices)
          do i = 1, numFilteredRows
            srcData(tempIndices(i))%label = intToStr(i)
          end do
          deallocate(tempIndices)

        case('expr')

!labelSortColName = generateNewColName(sourcesTab, (/nCutSortColName, nCutIndexColName, radiusColName/)) ! in ssclib/dal_aux.
labelSortColName = 'L_RESULT'

if (hasColumn(sourcesTab, labelSortColName)) then
!*** delete this column
call error('lResultColAlreadyExists', "slconv wants to make a column &
&in the source list to hold the result of its column arithmetic. &
&However it is not yet smart enough to choose an unused column name &
&- it insists on it being called L_RESULT. However there is already &
&a column of that name in the file.")
end if


          call initExprColGen(srcLabelSortExpression, labelSortColName, REAL32)
          call doExprColGen(sourcesTab, sourcesTab)

          allocate(sortColData(numFilteredRows))
          call readColDataToFixed(sourcesTab, labelSortColName, sortColData)
          sortColData = -sortColData
          allocate(tempIndices(numFilteredRows))
          tempIndices = (/(i, i=1,numFilteredRows)/)
          call quickSort(sortColData, tempIndices)
          do i = 1, numFilteredRows
            srcData(tempIndices(i))%label = intToStr(i)
          end do
          deallocate(sortColData, tempIndices)

        case('rownumber')
          do i = 1, numFilteredRows
!            srcData(srcData(i)%index)%label = intToStr(i)
            srcData(i)%label = intToStr(i)
          end do

        case default
call error('badLabelSortStyle', "The value of the --labelsortstyle &
&parameter was not recognized.")
        end select

      case('expr') ! srcLabelStyle
        if (hasColumn(sourcesTab, srcLabelExpression)) then
          labelStringCol = column(sourcesTab, srcLabelExpression)
          select case(dataType(labelStringCol))
          case(INTEGER8, INTEGER16, INTEGER32)
            allocate(int32ColData(numFilteredRows))
            call readColDataToFixed(labelStringCol, int32ColData)
            do i = 1, numFilteredRows
              srcData(i)%label = intToStr(int32ColData(i))
            end do
            deallocate(int32ColData)

          case(REAL32)
            allocate(real32ColData(numFilteredRows))
            call readColDataToFixed(labelStringCol, real32ColData)
            do i = 1, numFilteredRows
              write(srcData(i)%label, '(es13.6)') real32ColData(i)
            end do
            deallocate(real32ColData)

          case(REAL64)
            allocate(real64ColData(numFilteredRows))
            call readColDataToFixed(labelStringCol, real64ColData)
            do i = 1, numFilteredRows
              write(srcData(i)%label, '(es19.12)') real64ColData(i)
            end do
            deallocate(real64ColData)

          case(STRING)
!            call readColData(labelStringCol, srcData%label)
            do i = 1, numFilteredRows
!              srcData(i)%label = no0s(trim(stringCell(labelStringCol, i - 1)))
srcData(i)%label = trim(stringCell(labelStringCol, i - 1))
!*** is the no0s rendered unnecessary by the init to ''?
            end do

          case default
call error('badLabelColDataType',"The data type of the column specified &
&in parameter --labelexpression is not recognized.")
          end select

          call release(labelStringCol)
        else
call error('badLabelExpression',"The task has assumed that &
&--labelexpression contains just one column name. It has not found a &
&column of that name in the source list.")
        end if

      case default
call error('badLabelStyle', "The value of the --labelstyle parameter &
&was not recognized.")
      end select ! srcLabelStyle

      deallocate(tempRadii)!!, tempIndices)
    end if ! withSrcLabels

  end subroutine filterSourcesAndLoadSomeData

  !---------------------------------------------------------------------
  subroutine calcMainScale(srcData)
  !---------------------------------------------------------------------

    implicit none

    type(localSrcDataType), intent(in) :: srcData(:)

    type(DataSetT) :: imageSet
    type(ArrayT) :: imageArray
    integer(int32) ::&
      xsizePixels, ysizePixels
    real(single) ::&
      raSize, decSize,&
      xsizeArcsec, ysizeArcsec,&
squareXPixels(4), squareYPixels(4)
    integer(int32), dimension(:), pointer ::&
      dimVector
real(double) :: squareTanX(4), squareTanY(4)

    !-------------------------------------------------------------------
    ! Start

    if (withImageSet) then ! it is relatively easy: use 0.5% and 2.5%
    ! respectively of the largest image dimension as the min and max radii:
      imageSet = dataSet(imageSetName, READ)
      imageArray = array(imageSet, 0, READ)
      dimVector => dimensions(imageArray)
      xsizePixels = dimVector(0)
      ysizePixels = dimVector(1)
      call release(imageSet)

      if (radiusUnit .eq. 'pixels') then
        maxSize = max(xsizePixels, ysizePixels)

      else if (radiusUnit .eq. 'arcsec') then
!**** all this only works for RA---TAN, DEC--TAN stuff, and worldUnits = 'deg'!!! check these somewhere.
!        raSize  = xsizePixels * abs(wcs%x%pixelSizeWorld)
!        decSize = ysizePixels * abs(wcs%y%pixelSizeWorld)
!        xsizeArcsec = raSize * cos(PiOn180 * wcs%y%refWorld) * 3600.0
!        ysizeArcsec = decSize * 3600.0

squareXPixels = (/0.0, 0.0, real(xsizePixels), real(xsizePixels)/)
squareYPixels = (/0.0, real(ysizePixels), real(ysizePixels), 0.0/)
call pixelsToTan(squareXPixels, squareYPixels, wcs, squareTanX, squareTanY)
! in ssclib/wcs_aux
xsizeArcsec = 3600.0 * (maxval(squareTanX) - minval(squareTanX)) / PiOn180
ysizeArcsec = 3600.0 * (maxval(squareTanY) - minval(squareTanY)) / PiOn180
!*** all this is a bit rough and ready. A bit of skew could seriously stuff it up.

        maxSize = max(xsizeArcsec, ysizeArcsec)

      else
call error('badRadiusUnit',"Please notify the code developer.")
      end if

if (maxSize .le. 0.0) call error('slconvBug',"Please notify the &
&code developer.")

    else ! no imageset. Now it gets tricky. Best thing I can think of is to
    ! scale to the maximum size of the rectangle covered by the sources.
    ! Obviously this can only work if there are more than 2 of them
    ! separated in either RA or dec.
      raSize  = maxval(srcData%raDeg)  - minval(srcData%raDeg)
      decSize = maxval(srcData%decDeg) - minval(srcData%decDeg)

      xsizeArcsec = raSize * cos(PiOn180 * wcs%y%refWorld) * 3600.0
      ysizeArcsec = decSize * 3600.0
      maxSize = max(xsizeArcsec, ysizeArcsec)
      if (maxSize .le. 0.0) then
!**** warning??
        maxSize = lastResortFovDiameterArcsec
      end if
    end if

  end subroutine calcMainScale

  !---------------------------------------------------------------------
  subroutine saveAsDs9RegionFile(srcData)
  !---------------------------------------------------------------------

    implicit none

    type(localSrcDataType), intent(inout) :: srcData(:)

    character(255) ::&
      ds9RegionFileName
    integer ::&
      ioUnit, radiusInt, diameterInt
    integer(int32) ::&
      numSources,&
      numLastSource,&
      i, j
    character(20) ::&
      xStr, yStr, rStr, rStr2, jStr
    character(64) ::&
      strTextInReg
    real(single) ::&
      textYoffset  ! ,& textDecDeg
    real(double) :: tanX(size(srcData)), tanY(size(srcData))
!    logical(bool) :: someNintRadiiAreZero, allNintRadiiAreZero
    logical :: someNintRadiiAreZero, allNintRadiiAreZero

    !-------------------------------------------------------------------
    ! Start

    select case(outFileStyle)
    case('prefix')
      ds9RegionFileName = trim(outFileRoot)//".reg"
    case('whole')
      ds9RegionFileName = outFileName
    case default
call error('badOutFileStyle',"The value of the --outfilestyle parameter &
&was not recognized.")
    end select

    call getFreeIoUnit(ioUnit)
    open(ioUnit, file=ds9RegionFileName, status='REPLACE', access='SEQUENTIAL'&
    , form='FORMATTED')
    write(ioUnit,"(a)") '# Region file format: DS9 version 3.0'
!    write(ioUnit,"(a)") 'global color='//trim(colour)//' font="helvetica &
!    &10 normal" edit=0 move=0 delete=1 include=1 fixed=0'
    write(ioUnit,"(a)") 'global font="helvetica &
         &10 normal" edit=0 move=0 delete=1 include=1 fixed=0'

    numSources = size(srcData)
!    if (truncateSourceList) then
!      numLastSource = min(maxSourceNum, numSources)
!    else
      numLastSource = numSources
!    end if

    textYoffset = maxSize * textYOffsetFrac

    strTextInReg = ''		! i.e., if (.not. withSrcLabels)

    if (radiusUnit .eq. 'arcsec') then
      someNintRadiiAreZero = .false. ! default
      allNintRadiiAreZero  = .true.  ! default

      do i = 1, numLastSource
!        j = srcData(i)%index
j = i

        if (isNaN(srcData(j)%radius)) then
        ! (The function isNaN is in ssclib/ssc_misc.)
          jStr = intToStr(j, leftJustify=.true._bool)

!          call warning('badRadiusValue',"The source radius for source "&
!          //jStr//" was found to be NaN.")
      cycle
        end if

        radiusInt = nint(srcData(j)%radius)
        if (radiusInt <= 0) then
          someNintRadiiAreZero = .true.
          radiusInt = 0
        else
          allNintRadiiAreZero = .false.
        end if

        rStr = intToStr(radiusInt)
        diameterInt=2*radiusInt
        rStr2 = intToStr(diameterInt)

        if (withSrcLabels) then
          strTextInReg = ' text={'//trim(srcData(j)%label)//'}'
        end if

        select case(shape)
        case('circle')
          write(ioUnit, '(a, f10.6, a, f10.6, a)') 'fk5;circle('&
            , srcData(j)%raDeg, ',', srcData(j)%decDeg&
            , ','//trim(rStr)//'") # color='//trim(colour)&
            //trim(strTextInReg)
        case('box')
          write(ioUnit, '(a, f10.6, a, f10.6, a)') 'fk5;box('&
            , srcData(j)%raDeg, ',', srcData(j)%decDeg&
            , ','//trim(rStr2)//'",'//trim(rStr2)//'") # color='//trim(colour)&
            //trim(strTextInReg)
        case('diamond')
          write(ioUnit, '(a, f10.6, a, f10.6, a)') 'fk5;box('&
            , srcData(j)%raDeg, ',', srcData(j)%decDeg&
            , ','//trim(rStr2)//'",'//trim(rStr2)//'",45.0) # color='//trim(colour)&
            //trim(strTextInReg)
        case('ellipse')
          write(ioUnit, '(a, f10.6, a, f10.6, a)') 'fk5;ellipse('&
            , srcData(j)%raDeg, ',', srcData(j)%decDeg&
            , ','//trim(rStr2)//'",'//trim(rStr)//'") # color='//trim(colour)&
            //trim(strTextInReg)
        case('boxcircle')
          write(ioUnit, '(a, f10.6, a, f10.6, a)') 'fk5;box('&
            , srcData(j)%raDeg, ',', srcData(j)%decDeg&
            , ','//trim(rStr2)//'",'//trim(rStr2)//'") # color='//trim(colour)&
            //trim(strTextInReg)
          write(ioUnit, '(a, f10.6, a, f10.6, a)') 'fk5;circle('&
            , srcData(j)%raDeg, ',', srcData(j)%decDeg&
            , ','//trim(rStr)//'") # color='//trim(colour)
        case default
call error('badRegionShape', "The value of the shape parameter &
&was not recognized.")
        end select ! shape

        ! if (withSrcLabels) then
        !   textDecDeg = srcData(j)%decDeg + (textYoffset&
        !   + srcData(j)%radius) / 3600.0
        ! 
        !   write(ioUnit, '(a, f10.6, a, f10.6, a)') 'fk5;text('&
        !   , srcData(j)%raDeg, ',', textDecDeg&
        !   , ') # text={'//trim(srcData(j)%label)//'}'
        ! end if
      end do

      if (allNintRadiiAreZero) then
        call warning('allIntegerRadiiAreZero',"After conversion to integer, &
          &all of the radii are equal to zero. Maybe you want to make them &
          &bigger in the radius specification expression.")
      else if (someNintRadiiAreZero) then
        call warning('someIntegerRadiiAreZero',"After conversion to integer, &
          &some of the radii are equal to zero. Maybe you want to make them &
          &bigger in the radius specification expression.")
      end if

    else if (radiusUnit .eq. 'pixels') then
      call raDecToTan(srcData%raDeg*PiOn180, srcData%decDeg*PiOn180&
        , wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
        , tanX, tanY) ! in ssclib/coordinate

      call tanToPixels(wcs, tanX, tanY, srcData%xPixels, srcData%yPixels)
      ! in ssclib/wcs_aux
!**** all this only works for RA---TAN, DEC--TAN stuff!!! check it somewhere. AND check the worldUnits are both 'deg'.

      someNintRadiiAreZero = .false. ! default
      allNintRadiiAreZero  = .true.  ! default

      do i = 1, numLastSource
        j = i

        if (isNaN(srcData(j)%radius)) then
        ! (The function isNaN is in ssclib/ssc_misc.)
          jStr = intToStr(j, leftJustify=.true._bool)

!          call warning('badRadiusValue',"The source radius for source "&
!          //jStr//" was found to be NaN.")
      cycle
        end if

        xStr = intToStr(nint(srcData(j)%xPixels))
        yStr = intToStr(nint(srcData(j)%yPixels))

        radiusInt = nint(srcData(j)%radius)
        if (radiusInt <= 0) then
          someNintRadiiAreZero = .true.
          radiusInt = 0
        else
          allNintRadiiAreZero = .false.
        end if

        rStr = intToStr(radiusInt)

        if (withSrcLabels) then
          write(ioUnit, '(a)') 'image;circle('//trim(xStr)//','&
             //trim(yStr)//','//trim(rStr)//') # text={'//trim(srcData(j)%label)//'}'
        else
          write(ioUnit, '(a)') 'image;circle('//trim(xStr)//','&
             //trim(yStr)//','//trim(rStr)//')'
        end if

        ! if (withSrcLabels) then
        !   yStr = intToStr(nint(srcData(j)%yPixels + textYoffset&
        !   + srcData(j)%radius))
        ! 
        !   write(ioUnit, '(a)') 'image;text('//trim(xStr)//','&
        !   //trim(yStr)//') # text={'//trim(srcData(j)%label)//'}'
        ! end if
      end do

      if (allNintRadiiAreZero) then
        call warning('allIntegerRadiiAreZero',"After conversion to integer, &
          &all of the radii are equal to zero. Maybe you want to make them &
          &bigger in the radius specification expression.")
      else if (someNintRadiiAreZero) then
        call warning('someIntegerRadiiAreZero',"After conversion to integer, &
          &some of the radii are equal to zero. Maybe you want to make them &
          &bigger in the radius specification expression.")
      end if

    else
call error('badRadiusUnit',"Please notify the code developer.")
    end if	! if (radiusUnit .eq. 'arcsec') then

    close(ioUnit)

  end subroutine saveAsDs9RegionFile

  !---------------------------------------------------------------------
  subroutine saveAsGaiaRegionFile(srcData, sourcesSet)
  !---------------------------------------------------------------------
  ! See http://star-www.rl.ac.uk/star/docs/ssn75.html for info on the format
  ! of these tab-separated gaia tables.

    implicit none

    type(localSrcDataType), intent(inout) :: srcData(:)
    type(DataSetT), intent(in) :: sourcesSet

    character(255) :: gaiaRegionFileName
    integer :: ioUnit
    character(11) :: raStr(size(srcData))
    character(12) :: decStr(size(srcData))
    integer(int32) ::&
      i, j,&
      numSources,&
      numLastSource,&
      lastGaiaColNum,&
      startColNum,&
      colNum
    integer(int32), parameter ::&
      gaiaColValueSize = 20
    type(TableT) :: sourcesTab
    type(ColumnT) :: col
    character(lenGaiaColNames), allocatable ::&
      gaiaColNames(:)
    character(gaiaColValueSize), allocatable ::&
      gaiaColValues(:,:)
    character(1), parameter ::&
      tabChar=achar(9)
    integer(int32), allocatable :: int32ColData(:)
    real(single),   allocatable :: real32ColData(:)
    real(double),   allocatable :: real64ColData(:)
    integer(int16), allocatable :: maxLenTrim(:)
!    character(gaiaColValueSize), allocatable :: stringColData(:)

    !-------------------------------------------------------------------
    ! Start

    numSources = size(srcData)
!    if (truncateSourceList) then
!      numLastSource = min(maxSourceNum, numSources)
!    else
      numLastSource = numSources
!    end if

!    if (srcLabelStyle .eq. 'none') then
if (.not. withSrcLabels) then
      lastGaiaColNum = numMandatoryGaiaCols + size(extraGaiaColNames)
      allocate(gaiaColNames(lastGaiaColNum))
      allocate(gaiaColValues(numLastSource, lastGaiaColNum))

      gaiaColNames(:numMandatoryGaiaCols) = mandatoryGaiaColNames
      gaiaColNames(numMandatoryGaiaCols + 1:) = extraGaiaColNames

    else
      lastGaiaColNum = numMandatoryGaiaCols + size(extraGaiaColNames) + 1
      allocate(gaiaColNames(lastGaiaColNum))
      allocate(gaiaColValues(numLastSource, lastGaiaColNum))

      gaiaColNames(:numMandatoryGaiaCols) = mandatoryGaiaColNames
      gaiaColNames(numMandatoryGaiaCols + 1) = 'label'
      gaiaColNames(numMandatoryGaiaCols + 2:) = extraGaiaColNames
    end if

    gaiaColValues = ''

    call convertRa( srcData%raDeg,  raStr)
    call convertDec(srcData%decDeg, decStr)

    if (radiusUnit .eq. 'arcsec') then ! change it to degrees - gaia has no
    ! valid arcsec unit.
      srcData%radius = srcData%radius / 3600.0
    end if

    ! Fill the temporary gaia tabel with left-justified, space-padded
    ! string values:
    do i = 1, numLastSource
!      j = srcData(i)%index
j = i

      gaiaColValues(i, 1) = intToStr(i)	! Id
      gaiaColValues(i, 2) = raStr(j)	! ra
      gaiaColValues(i, 3) = decStr(j)	! dec
      write(gaiaColValues(i, 4), '(es13.6)') srcData(j)%radius
!      call smartRealToStringConversion(srcData(j)%radius&
!      , gaiaColValues(i, 4), status)
!!*** call error on bad status
    end do

!    if (srcLabelStyle .eq. 'none') then
if (.not. withSrcLabels) then
      startColNum = numMandatoryGaiaCols + 1
    else
      do i = 1, numLastSource
!        j = srcData(i)%index
j = i
        gaiaColValues(i, numMandatoryGaiaCols + 1) = srcData(j)%label
      end do
      startColNum = numMandatoryGaiaCols + 2
    end if

    sourcesTab = table(sourcesSet, srcListTabName)

    do i = 1, size(extraGaiaColNames)
      if (.not. hasColumn(sourcesTab, extraGaiaColNames(i))) then
        write(errstr, '(a, a, a)') "You have asked that values from a &
          &column called ", trim(extraGaiaColNames(i)), " be written to &
          &the GAIA output. But there is no such column in the source list &
          &table. You will have to remove this column name from the &
          &--extragaiacolumns parameter before the task will run with the &
          &present source list."

call error('badExtraGaiaColName')
      end if
    end do

    do colNum = startColNum, lastGaiaColNum
      col = column(sourcesTab, gaiaColNames(colNum))
      select case(dataType(col))
      case(INTEGER8, INTEGER16, INTEGER32)
        allocate(int32ColData(numLastSource))
        call readColDataToFixed(col, int32ColData)
        do i = 1, numLastSource
!          j = srcData(i)%index
j = i
          gaiaColValues(i, colNum) = intToStr(int32ColData(j))
        end do
        deallocate(int32ColData)

      case(REAL32)
        allocate(real32ColData(numLastSource))
        call readColDataToFixed(col, real32ColData)
        do i = 1, numLastSource
!          j = srcData(i)%index
j = i
          write(gaiaColValues(i, colNum), '(es13.6)') real32ColData(j)
!          call smartRealToStringConversion(real32ColData(j)&
!          , gaiaColValues(i, colNum), status)
!*** call error on bad status
        end do
        deallocate(real32ColData)

      case(REAL64)
        allocate(real64ColData(numLastSource))
        call readColDataToFixed(col, real64ColData)
        do i = 1, numLastSource
!          j = srcData(i)%index
j = i
          write(gaiaColValues(i, colNum), '(es19.12)') real64ColData(j)
!          call smartRealToStringConversion(real64ColData(j)&
!          , gaiaColValues(i, colNum), status)
!*** call error on bad status
        end do
        deallocate(real64ColData)

      case(STRING)
!        allocate(stringColData(numLastSource))
        do i = 1, numLastSource
!          j = srcData(i)%index
j = i
!          gaiaColValues(i, colNum) = no0s(trim(stringCell(col, j - 1)))
gaiaColValues(i, colNum) = trim(stringCell(col, j - 1))
!*** is the no0s rendered unnecessary by the init to ''?
        end do

      case default
        write(errstr, '(a, a, a)') "The data type of the "&
          , trim(gaiaColNames(colNum)), " column was not recognized."

call error('badExtraGaiaColDataType')
      end select

      call release(col)
    end do

    allocate(maxLenTrim(lastGaiaColNum))
    do j = 1, lastGaiaColNum
      maxLenTrim(j) = maxval(len_trim(gaiaColValues(:,j)))
    end do

    ! Open the file and write the data:

    select case(outFileStyle)
    case('prefix')
      gaiaRegionFileName = trim(outFileRoot)//".gaia"
    case('whole')
      gaiaRegionFileName = outFileName
    case default
call error('badOutFileStyle',"The value of the --outfilestyle parameter &
&was not recognized.")
    end select

    call getFreeIoUnit(ioUnit)
    open(ioUnit, file=gaiaRegionFileName, status='REPLACE'&
    , access='SEQUENTIAL', form='FORMATTED')

    write(ioUnit,"(a)") 'serv_type: local'

!    if (srcLabelStyle .eq. 'none') then
if (.not. withSrcLabels) then
      if (radiusUnit .eq. 'pixels') then
        write(ioUnit,"(a)") 'symbol: {radius} &
        &{circle '//trim(colour)//'} {$radius image}'
      else if (radiusUnit .eq. 'arcsec') then
        write(ioUnit,"(a)") 'symbol: {radius} &
        &{circle '//trim(colour)//'} {$radius "deg J2000"}'
      else
call error('badRadiusUnit',"Please notify the code developer.")
      end if
    else
      if (radiusUnit .eq. 'pixels') then
        write(ioUnit,"(a)") 'symbol: {label radius} &
        &{circle '//trim(colour)//' "" "" $label} {$radius image}'
      else if (radiusUnit .eq. 'arcsec') then
        write(ioUnit,"(a)") 'symbol: {label radius} &
        &{circle '//trim(colour)//' "" "" $label} {$radius "deg J2000"}'
      else
call error('badRadiusUnit',"Please notify the code developer.")
      end if
    end if

    write(ioUnit,"(a)") '' ! linebreak just to make it look a bit better.

    ! Print the column names row and the tab-separated dashes beneath it:
    do i = 1, lastGaiaColNum - 1
      write(ioUnit,"(a)", advance='NO') trim(gaiaColNames(i))//tabChar
    end do
    write(ioUnit,"(a)") trim(gaiaColNames(lastGaiaColNum))
    do i = 1, lastGaiaColNum - 1
      write(ioUnit,"(a)", advance='NO') repeat('-'&
      , len_trim(gaiaColNames(i)))//tabChar
    end do
    write(ioUnit,"(a)") repeat('-'&
    , len_trim(gaiaColNames(lastGaiaColNum)))

    ! Now print the table:
    do i = 1, numLastSource
      do j = 1, lastGaiaColNum - 1
        write(ioUnit,"(a)", advance='NO') gaiaColValues(i, j)(:maxLenTrim(j))&
        //tabChar
      end do
      write(ioUnit,"(a)") gaiaColValues(i&
      , lastGaiaColNum)(:maxLenTrim(lastGaiaColNum))
    end do

    close(ioUnit)

  end subroutine saveAsGaiaRegionFile

  !---------------------------------------------------------------------
  subroutine convertRa(raDeg,  raStr)
  !---------------------------------------------------------------------
!*** now in ssclib/sscmisc as raDegToStr.

    implicit none

    real(double), intent(in) :: raDeg(:)
    character(11), intent(out) :: raStr(size(raDeg))

    real(single) :: raHours, raMin, raSec
    integer(int16) :: raHoursInt, raMinInt, i
    character(5) :: raSecStr

    !-------------------------------------------------------------------
    ! Start

    do i = 1, size(raDeg)
      raHours = raDeg(i) / 15.0
      raHoursInt = floor(raHours)
      raMin = (raHours - raHoursInt) * 60.0
      raMinInt = floor(raMin)
      raSec = (raMin - raMinInt) * 60.0

      if (raSec .lt. 1.0) then
        write(raSecStr, '(a2, f3.2)') '00', raSec
      else if (raSec .lt. 10.0) then
        write(raSecStr, '(a1, f4.2)') '0', raSec
      else
        write(raSecStr, '(f5.2)') raSec
      end if

      write(raStr(i), '(i2.2, a1, i2.2, a1, a5)') raHoursInt, ':', raMinInt&
      , ':', raSecStr
    end do

  end subroutine convertRa

  !---------------------------------------------------------------------
  subroutine convertDec(decDeg, decStr)
  !---------------------------------------------------------------------
!*** now in ssclib/sscmisc as decDegToStr.

    implicit none

    real(double), intent(in) :: decDeg(:)
    character(12), intent(out) :: decStr(size(decDeg))

    real(single) :: absDecDeg, decMin, decSec
    integer(int16) :: absDecDegInt, decMinInt, i
    character(1) :: decSignStr
    character(5) :: decSecStr

    !-------------------------------------------------------------------
    ! Start

    do i = 1, size(decDeg)
      if(decDeg(i) .lt. 0.d0)then
        decSignStr = '-'
      else
        decSignStr = '+'
      endif

      absDecDeg = abs(decDeg(i))
      absDecDegInt = floor(absDecDeg)
      decMin = (absDecDeg - absDecDegInt) * 60.0
      decMinInt = floor(decMin)
      decSec = (decMin - decMinInt) * 60.0

      if (decSec .lt. 1.0) then
        write(decSecStr, '(a2, f3.2)') '00', decSec
      else if (decSec .lt. 10.0) then
        write(decSecStr, '(a1, f4.2)') '0', decSec
      else
        write(decSecStr, '(f5.2)') decSec
      end if

      write(decStr(i), '(a1, i2.2, a1, i2.2, a1, a5)') decSignStr, absDecDegInt&
      , ':', decMinInt, ':', decSecStr
    end do

  end subroutine convertDec

!  !---------------------------------------------------------------------
!  subroutine raStrToDeg(raStr, raDeg, status)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    !-------------------------------------------------------------------
!    ! Start
!
!  end subroutine raStrToDeg
!
!  !---------------------------------------------------------------------
!  subroutine decStrToDeg(decStr, decDeg, status)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    !-------------------------------------------------------------------
!    ! Start
!
!  end subroutine decStrToDeg

!  !---------------------------------------------------------------------
!  subroutine smartRealToStringConversion(realValue, string, status)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real(double), intent(in) :: realValue
!    character(*), intent(out) :: string
!    integer, intent(out) :: status
!
!    integer :: intValue
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    status = 0
!    string = ''
!
!    strLength = len(string)
!
!    ! Find out first if the value is effectively an integer:
!    if (nint(realValue) .eq. realValue&
!    .and. 1 + log10(realValue) .le. strLength) then
!      string = intToStr(nint(realValue))
!  return
!    end if
!
!    ! Either it wasn't an integer, or it was, but was too big to fit in the field. Check now to see if it is a non-integer that will fit in the field in a reasonable way without conversion to exponential format:
!    if (nint(realValue) .eq. realValue .and. ) then
!  return
!    end if
!
!  end subroutine smartRealToStringConversion


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module slconv_mod
